/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/* i/o layering support */

#ifndef AP_IOL_H
#define AP_IOL_H

#include "apr_general.h" /* For ap_s?size_t */
#include "apr_network_io.h" /* For ap_hdtr_t */
#include "apr_errno.h" /* For ap_status_t and the APR_errnos */
#include "ap_config.h" /* For ap_status_t and the APR_errnos */

typedef struct ap_iol ap_iol;
typedef struct ap_iol_methods ap_iol_methods;

typedef enum {
    AP_IOL_TIMEOUT
} ap_iol_option;

/*
    write, writev, read guarantee they won't return APR_EINTR.

    If no error occurs, they return APR_SUCCESS.

    If timeout is negative, they may block forever, and they still don't
    guarantee they'll write/read the full length.

    If timeout is 0, they will never block, and will return APR_EAGAIN
    in the event that they can't read/write any bytes.

    If the timeout is positive, they will block up to the specified number
    of seconds.  If the read/write can't be completed in that time,
    APR_ETIMEDOUT will be returned.

    Regardless of the return value, bytes may have been read or written.
    (For example, APR_EOF may be returned along with 100 bytes from read.)

    XXX: timeouts should probably be redefined to be in microseconds to match APR.

    Just to drive the point home again -- none of these functions guarantee
    they will read/write the entire length specified.

    Note: timeout == 0 is non-blocking operation... which requires a lot
    of care if you're implementing filters.  It may mean you have to keep
    state from one call to the next.  The chunking filter will have an
    example of this.  If you're not sure how to handle the state sitatuation
    you will want to return APR_EINVAL when the timeout is set to 0.  This may
    break various things... hopefully we'll be able to recover gracefully.

    Note: implementation tip: if your system doesn't have writev(), you
    still have to implement a writev() method.  However, since the
    iol writev() is allowed to return a partial write, you can essentially
    treat all writev()s as write(fd, vec[0].iov_base, vec[0].iov_len),
    and the layers above will loop over the entire vector for you.
*/

/**
 * @package IO Layering library
 */

struct ap_iol_methods {
    apr_status_t (*close)(ap_iol *fd);
    apr_status_t (*write)(ap_iol *fd, const char *buf, apr_size_t len,
                         apr_ssize_t *nbytes);
    apr_status_t (*writev)(ap_iol *fd, const struct iovec *vec, apr_size_t nvec,
                          apr_ssize_t *nbytes);
    apr_status_t (*read)(ap_iol *fd, char *buf, apr_size_t len,
                        apr_ssize_t *nbytes);
    apr_status_t (*setopt)(ap_iol *fd, ap_iol_option opt, const void *value);
    apr_status_t (*getopt)(ap_iol *fd, ap_iol_option opt, void *value);
    apr_status_t (* sendfile)(ap_iol *fd, apr_file_t * file, apr_hdtr_t * hdtr, 
                             apr_off_t * offset, apr_size_t * len, 
                             apr_int32_t flags);
    apr_status_t (*shutdown)(ap_iol *fd, int how);
    /* TODO: accept, connect, ... */
};

/* it is expected that each io layer will extend this structure as they
    require; they must only ensure that the methods pointer remains in
    the indicated position. */
/**
 * ap_iol structure
 * This is an abstraction of Input/Output types and functions.  This
 * structure is composed of a list of function pointers.  Those functions
 * should only be accessed through the macros provided below.  The function
 * pointers.
 * @tip Not all of the IOL types have all of the IOL functions.  For example
 * a file IOL does not support sendfile or shutdown, so those function pointers
 * are NULL.
 * <PRE>
 * apr_status_t (*close)(ap_iol *fd);
 * apr_status_t (*write)(ap_iol *fd, const char *buf, apr_size_t len, apr_ssize_t *nbytes);
 * apr_status_t (*writev)(ap_iol *fd, const struct iovec *vec, apr_size_t nvec, apr_ssize_t *nbytes);
 * apr_status_t (*read)(ap_iol *fd, char *buf, apr_size_t len, apr_ssize_t *nbytes);
 * apr_status_t (*setopt)(ap_iol *fd, ap_iol_option opt, const void *value);
 * apr_status_t (*getopt)(ap_iol *fd, ap_iol_option opt, void *value);
 * apr_status_t (* sendfile)(ap_iol *fd, apr_file_t *file, apr_hdtr_t *hdtr, apr_off_t *offset, apr_size_t *len, apr_int32_t flags);
 * apr_status_t (*shutdown)(ap_iol *fd, int how);
 * </PRE>
 */
struct ap_iol {
    const ap_iol_methods *methods;
};

/**
 * Macro to close the specified iol
 * @param iol The iol to close
 * @deffunc apr_status_t iol_close(ap_iol *iol)
 */
#define iol_close(iol)	((iol)->methods->close((iol)))
/**
 * Macro to write a buffer to the iol
 * @param iol The iol to write to.
 * @param buf The buffer to write
 * @param len The length of the buffer
 * @param nbytes The number of bytes actually written
 * @tip This function is not garuanteed to write the whole buffer.  It will
 * return APR_SUCCESS if any bytes were written without errors.
 * @deffunc apr_status_t iol_write(ap_iol *iol, const char *buf, apr_size_t len, apr_ssize_t *nbytes)
 */
#define iol_write(iol, a, b, c) ((iol)->methods->write((iol), (a), (b), (c)))
/**
 * Macro to write an iovec to the iol
 * @param iol The iol to write to.
 * @param vec The iovec to write
 * @param nvec The number of elements in the iovec
 * @param nbytes The number of bytes actually written
 * @tip This function is not garuanteed to write the whole buffer.  It will
 * return APR_SUCCESS if any bytes were written without errors.
 * @deffunc apr_status_t iol_writev(ap_iol *iol, const struct iovec *vec, apr_size_t nvec, apr_ssize_t *nbytes)
 */
#define iol_writev(iol, a, b, c) ((iol)->methods->writev((iol), (a), (b), (c)))
/**
 * Macro to read a buffer from an iol
 * @param iol The iol to write to.
 * @param buf The buffer to read into
 * @param len The size of the buffer
 * @param nbytes The number of bytes actually read
 * @tip This function is not garuanteed to read the number of bytes specified.
 * It will return APR_SUCCESS if any bytes were written without errors.
 * @deffunc apr_status_t iol_read(ap_iol *iol, char *buf, apr_size_t len, apr_ssize_t *nbytes)
 */
#define iol_read(iol, a, b, c) ((iol)->methods->read((iol), (a), (b), (c)))
/**
 * Macro to set an iol option
 * @param iol The iol set the option for.
 * @param opt The option to set
 * @param value The value to set the option to.
 * @tip Currently, the only option recognized is AP_IOL_TIMEOUT.  This sets
 * the timeout value for all read/write operations.  The value should be
 * specified in seconds.
 * @deffunc apr_status_t iol_setopt(ap_iol *iol, ap_iol_option opt, const void *value)
 */
#define iol_setopt(iol, a, b) ((iol)->methods->setopt((iol), (a), (b)))
/**
 * Macro to get an iol option
 * @param iol The iol get the option for.
 * @param opt The option to get
 * @param value The value to get the option to.
 * @tip Currently, the only option recognized is AP_IOL_TIMEOUT.  This gets
 * the current timeout value for all read/write operations.  The value 
 * is specified in seconds.
 * @deffunc apr_status_t iol_setopt(ap_iol *iol, ap_iol_option opt, void *value)
 */
#define iol_getopt(iol, a, b) ((iol)->methods->getopt((iol), (a), (b)))
/**
 * Macro to send a file through a given IOL
 * @param iol The iol send the file through
 * @param file The file to send 
 * @param hdtr Any headers and trailers to send before the file
 * @param offset Place in the file to start sending the file
 * @param len The amount of the file to send
 * @param flags Flags mapped to OS specific flags.
 * @deffunc apr_status_t iol_sendfile(ap_iol *iol, apr_file_t *file, apr_hdtr_t *hdtr, apr_off_t *offset, apr_size_t *len, apr_int32_t flags)
 */
#define iol_sendfile(iol, a, b, c, d, e) ((iol)->methods->sendfile((iol), (a), (b), (c), (d), (e)))
/**
 * Macro to shutdown a given iol
 * @param iol The iol to shutdown
 * @param how How to shutdown the iol.
 * @deffunc apr_status_t iol_shutdown(ap_iol *iol, int how)
 */
#define iol_shutdown(iol, a) ((iol)->methods->shutdown((iol), (a)))

/* the file iol */
/**
 * Create an iol out of a file structure
 * @param file The file to insert into the iol
 */
ap_iol *ap_create_file_iol(apr_file_t *file);

/* the socket iol */
/**
 * Create an iol out of a socket structure
 * @param The pool to allocate the iol structure out of
 * @param sock The socket to insert into the iol
 * @deffunc ap_iol *ap_iol_attach_socket(apr_pool_t *p, apr_socket_t *sock)
 */
API_EXPORT(ap_iol *) ap_iol_attach_socket(apr_pool_t *p, apr_socket_t *sock);

/**
 * Get a socket out of an iol
 * @param iol The iol to get the socket from
 * @return The socket the iol points to.
 * @deffunc ap_socket_t *ap_iol_get_socket(ap_iol *iol)
 */
API_EXPORT(apr_socket_t *) ap_iol_get_socket(ap_iol *iol);

#endif
