/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* everything garbagearific */

#include "mod_cache.h"
#ifdef WIN32
#include <sys/utime.h>
#else
#include <utime.h>
#endif /* WIN32 */
#include <pthread.h>
#include "http_log.h"
#include "http_main.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

/* I use this instead of HUGE_STRING_LEN because I can probably come up with a better calculated
   number, and I'd rather mark with this so I can come back and FIXME. */
#define LARGE_FILE_NAME HUGE_STRING_LEN

/* Poor man's 61 bit arithmetic */
typedef struct {
    long lower;    /* lower 30 bits of result */
    long upper; /* upper 31 bits of result */
} long61_t;

struct gc_ent {
    unsigned long int len;
    apr_time_t expire;
    char base[LARGE_FILE_NAME];
};

/* FIXME: The block size can be different on a `per file system' base.
 * This would make automatic detection highly OS specific.
 * In the GNU fileutils code for du(1), you can see how complicated it can
 * become to detect the block size. And, with BSD-4.x fragments, it
 * it even more difficult to get precise results.
 * As a compromise (and to improve on the incorrect counting of cache
 * size on byte level, omitting directory sizes entirely, which was
 * used up to apache-1.3b7) we're rounding to multiples of 512 here.
 * Your file system may be using larger blocks (I certainly hope so!)
 * but it will hardly use smaller blocks.
 * (So this approximation is still closer to reality than the old behavior).
 * The best solution would be automatic detection, the next best solution
 * IMHO is a sensible default and the possibility to override it.
 */

#define ROUNDUP2BLOCKS(_bytes) (((_bytes)+block_size-1) & ~(block_size-1))
static long block_size = 512;    /* this must be a power of 2 */
static long61_t curbytes, cachesize;
static apr_time_t garbage_now, garbage_expire;
static pthread_mutex_t garbage_mutex = PTHREAD_MUTEX_INITIALIZER;


static int sub_garbage_coll(ap_cache_handle_t *r, apr_array_header_t *files,
                const char *cachedir, const char *cachesubdir);
static void help_cache_garbage_coll(struct ap_cache_handle_t *r);
static int should_cache_garbage_coll(struct ap_cache_handle_t *r);
#if !defined(WIN32) && !defined(MPE) && !defined(OS2)
static void detached_cache_garbage_coll(ap_cache_handle_t *r);
#endif

apr_status_t file_cache_gc(ap_cache_handle_t *r)
{
    static int inside = 0;

    (void) pthread_mutex_lock(&garbage_mutex);
    if (inside == 1) {
        (void) pthread_mutex_unlock(&garbage_mutex);
        return APR_ENOLOCK;    /* @@@FIXME */
    }
    else
    inside = 1;
    (void) pthread_mutex_unlock(&garbage_mutex);

    if (should_cache_garbage_coll(r))
#if !defined(WIN32) && !defined(MPE) && !defined(OS2) 
        detached_cache_garbage_coll(r);
#else
        help_cache_garbage_coll(r);
#endif

    (void) pthread_mutex_lock(&garbage_mutex);
    inside = 0;
    (void) pthread_mutex_unlock(&garbage_mutex);
    return APR_SUCCESS;
}


static void
add_long61 (long61_t *accu, long val)
{
    /* Add in lower 30 bits */
    accu->lower += (val & 0x3FFFFFFFL);
    /* add in upper bits, and carry */
    accu->upper += (val >> 30) + ((accu->lower & ~0x3FFFFFFFL) != 0L);
    /* Clear carry */
    accu->lower &= 0x3FFFFFFFL;
}

static void
sub_long61 (long61_t *accu, long val)
{
    int carry = (val & 0x3FFFFFFFL) > accu->lower;
    /* Subtract lower 30 bits */
    accu->lower = accu->lower - (val & 0x3FFFFFFFL) + ((carry) ? 0x40000000 : 0);
    /* add in upper bits, and carry */
    accu->upper -= (val >> 30) + carry;
}

/* Compare two long61's:
 * return <0 when left < right
 * return  0 when left == right
 * return >0 when left > right
 */
static long
cmp_long61 (long61_t *left, long61_t *right)
{
    return (left->upper == right->upper) ? (left->lower - right->lower)
                     : (left->upper - right->upper);
}

/* Compare two gc_ent's, sort them by expiration date */
static int gcdiff(const void *ap, const void *bp)
{
    const struct gc_ent *a = (const struct gc_ent * const) ap;
    const struct gc_ent *b = (const struct gc_ent * const) bp;

    if (a->expire > b->expire)
        return 1;
    else if (a->expire < b->expire)
        return -1;
    else
        return 0;
}

#if !defined(WIN32) && !defined(MPE) && !defined(OS2)
static void detached_cache_garbage_coll(ap_cache_handle_t *r)
{
    pid_t pid;
    int status;
    pid_t pgrp;

#if 0
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
             "cache: Guess what; we fork() again...");
#endif
    switch (pid = fork()) {
    case -1:
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
             "cache: fork() for cache cleanup failed");
        return;

    case 0:    /* Child */

        /* close all sorts of things, including the socket fd */
        apr_cleanup_for_exec();

        /* Fork twice to disassociate from the child */
        switch (pid = fork()) {
        case -1:
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
             "cache: fork(2nd) for cache cleanup failed");
            exit(1);

        case 0:    /* Child */
            /* The setpgrp() stuff was snarfed from http_main.c */
#ifndef NO_SETSID
            if ((pgrp = setsid()) == -1) {
            perror("setsid");
            ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO,  0, NULL, "%s: setsid failed",
                     ap_server_argv0);
            exit(1);
            }
#elif defined(NEXT) || defined(NEWSOS)
            if (setpgrp(0, getpid()) == -1 || (pgrp = getpgrp(0)) == -1) {
            perror("setpgrp");
            ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, NULL, "%S: setpgrp or getpgrp failed",
                     ap_server_argv0);
            exit(1);
            }
#else
            if ((pgrp = setpgrp(getpid(), 0)) == -1) {
            perror("setpgrp");
            ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                                     NULL, "%s: setpgrp failed",
                     ap_server_argv0);
            exit(1);
            }
#endif
            help_cache_garbage_coll(r);
            exit(0);

        default:    /* Father */
            /* After grandson has been forked off, */
            /* there's nothing else to do. */
            exit(0);            
        }
    default:
        /* Wait until grandson has been forked off */
        /* (without wait we'd leave a zombie) */
        waitpid(pid, &status, 0);
        return;
    }
}
#endif /* ndef WIN32 */

#define DOT_TIME "/.time"    /* marker */

static int should_cache_garbage_coll(ap_cache_handle_t *r)
{
    cache_server_conf *conf = (cache_server_conf *) ap_get_module_config(r->server->module_config, &cache_module);
    const char *cachedir = conf->root;
    char *filename;
    struct stat buf;
    int timefd;
    apr_time_t every = conf->gcinterval;
    static apr_time_t lastcheck = BAD_DATE;         /* static (per-process) data!!! */

    if (cachedir == NULL || every == -1)
        return 0;

    filename = apr_palloc(r->pool, strlen(cachedir) + strlen( DOT_TIME ) +1);

    garbage_now = apr_now();
    /* Usually, the modification time of <cachedir>/.time can only increase.
     * Thus, even with several child processes having their own copy of
     * lastcheck, if time(NULL) still < lastcheck then it's not time
     * for GC yet.
     */
    if (garbage_now != -1 && lastcheck != BAD_DATE && garbage_now < lastcheck + every)
        return 0;


    strcpy(filename, cachedir);
    strcat(filename, DOT_TIME);

    /* At this point we have a bit of an engineering compromise. We could either
     * create and/or mark the .time file  (prior to the fork which might
     * fail on a resource issue) or wait until we are safely forked. The
     * advantage of doing it now in this process is that we get some
     * usefull live out of the global last check variable. (XXX which
     * should go scoreboard IMHO.) Note that the actual counting is 
     * at a later moment.
     */
    if (stat(filename, &buf) == -1) {        /* does not exist */
        if (errno != ENOENT) {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                         "cache: stat(%s)", filename);
            return 0;
        }
        if ((timefd = creat(filename, 0666)) == -1) {
            if (errno != EEXIST)
                ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                             "cache: creat(%s)", filename);
            else
                lastcheck = garbage_now;        /* someone else got in there */
            return 0;
        }
        close(timefd);
    }
    else {
        lastcheck = buf.st_mtime * APR_USEC_PER_SEC;        /* save the time */
        if (garbage_now < lastcheck + every) {
            return 0;
        }
        if (utime(filename, NULL) == -1)
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                         "cache: utimes(%s)", filename);
    }

    return 1;
}

static void help_cache_garbage_coll(ap_cache_handle_t *r)
{
    const char *cachedir;
    cache_server_conf *conf = (cache_server_conf *) ap_get_module_config(r->server->module_config, &cache_module);
    apr_array_header_t *files;
    struct gc_ent *fent;
    char filename[LARGE_FILE_NAME];
    int i;

    cachedir = conf->root;
    /* configured size is given in kB. Make it bytes, convert to long61_t: */
    cachesize.lower = cachesize.upper = 0;
    add_long61(&cachesize, conf->space << 10);

    files = apr_make_array(r->pool, 100, sizeof(struct gc_ent));
    curbytes.upper = curbytes.lower = 0L;

    sub_garbage_coll(r, files, cachedir, "/");

    if (cmp_long61(&curbytes, &cachesize) < 0L) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r->server,
                     "cache GC: Cache is %ld%% full (nothing deleted)",
                     (long)(((curbytes.upper<<20)|(curbytes.lower>>10))*100/conf->space));
        return;
    }

    /* sort the files we found by expiration date */
    qsort(files->elts, files->nelts, sizeof(struct gc_ent), gcdiff);

    for (i = 0; i < files->nelts; i++) {
        fent = &((struct gc_ent *) files->elts)[i];
        sprintf(filename, "%s%s%s", cachedir, fent->base, CACHE_HEADER_SUFFIX);
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r->server,
                     "GC Unlinking %s (expiry %ld, garbage_now %ld)", filename, (long)fent->expire, (long)garbage_now);
#if TESTING
        ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, NULL, 
                     "Would unlink %s", filename);
#else
        if (unlink(filename) == -1) {
            if (errno != ENOENT)
                ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                             "cache gc: unlink(%s)", filename);
        }
#endif

        sprintf(filename, "%s%s%s", cachedir, fent->base, CACHE_DATA_SUFFIX);
#if TESTING
        ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, NULL, 
                     "Would unlink %s", filename);
#else
        if (unlink(filename) == -1) {
            if (errno != ENOENT)
                ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                             "cache gc: unlink(%s)", filename);
        }
#endif

        sub_long61(&curbytes, ROUNDUP2BLOCKS(fent->len));
        if (cmp_long61(&curbytes, &cachesize) < 0)
            break;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r->server,
                 "cache GC: Cache is %ld%% full (%d deleted)",
                 (long)(((curbytes.upper<<20)|(curbytes.lower>>10))*100/conf->space), i);
}

static int sub_garbage_coll(ap_cache_handle_t *r, apr_array_header_t *files,
              const char *cachebasedir, const char *cachesubdir)
{
    char line[27];
    char cachedir[HUGE_STRING_LEN];
    struct stat buf;
    int fd, i, hdrfile;
    DIR *dir;
#if defined(NEXT) || defined(WIN32)
    struct DIR_TYPE *ent;
#else
    struct dirent *ent;
#endif
    struct gc_ent *fent;
    int nfiles = 0;
    const int basedirlen = strlen(cachebasedir);
    char *filename, basename[LARGE_FILE_NAME], *s;

    apr_snprintf(cachedir, sizeof(cachedir), "%s%s", cachebasedir, cachesubdir);
    filename = apr_palloc(r->pool, strlen(cachedir) + HASH_LEN + 2);
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                 "GC Examining directory %s", cachedir);
    dir = opendir(cachedir);
    if (dir == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                     "cache gc: opendir(%s)", cachedir);
        return 0;
    }

    while ((ent = readdir(dir)) != NULL) {
        if (ent->d_name[0] == '.')
            continue;
        sprintf(filename, "%s%s", cachedir, ent->d_name);
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                     "GC Examining file %s", filename);
/* is it a temporary file? */
        if (strncmp(ent->d_name, "tmp", 3) == 0) {
/* then stat it to see how old it is; delete temporary files > 1 day old */
            if (stat(filename, &buf) == -1) {
                if (errno != ENOENT)
                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                                 "cache gc: stat(%s)", filename);
            }
            else if (garbage_now != -1 &&
                     (buf.st_atime*APR_USEC_PER_SEC) < garbage_now - MSEC_ONE_DAY &&
                     (buf.st_mtime*APR_USEC_PER_SEC) < garbage_now - MSEC_ONE_DAY) {
                ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                             "GC unlink %s", filename);
                ap_log_error(APLOG_MARK, APLOG_INFO|APLOG_NOERRNO, 0, r->server,
                             "cache gc: deleting orphaned cache file %s", filename);
#if TESTING
                ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                             NULL, "Would unlink %s", filename);
#else
                unlink(filename);
#endif
            }
            continue;
        }
        ++nfiles;
/* under OS/2 use dirent's d_attr to identify a diretory */
#ifdef OS2
/* is it a directory? */
        if (ent->d_attr & A_DIR) {
            char newcachedir[HUGE_STRING_LEN];
            apr_snprintf(newcachedir, sizeof(newcachedir),
                        "%s%s/", cachesubdir, ent->d_name);
            if (!sub_garbage_coll(r, files, cachebasedir, newcachedir)) {
                apr_snprintf(newcachedir, sizeof(newcachedir),
                            "%s%s", cachedir, ent->d_name);
#if TESTING
                ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                             NULL, "Would remove directory %s", newcachedir);
#else
                rmdir(newcachedir);
#endif
                --nfiles;
            }
            continue;
        }
#endif

/* read the file */
        fd = open(filename, O_RDONLY | O_BINARY);
        if (fd == -1) {
            if (errno != ENOENT)
                ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                             "cache gc: open(%s)", filename);
            continue;
        }
        if (fstat(fd, &buf) == -1) {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                         "cache gc: fstat(%s)", filename);
            close(fd);
            continue;
        }

/* In OS/2 this has already been done above */
#ifndef OS2
        if (S_ISDIR(buf.st_mode)) {
            char newcachedir[HUGE_STRING_LEN];
            close(fd);
            apr_snprintf(newcachedir, sizeof(newcachedir),
                        "%s%s/", cachesubdir, ent->d_name);
            if (!sub_garbage_coll(r, files, cachebasedir, newcachedir)) {
                apr_snprintf(newcachedir, sizeof(newcachedir),
                            "%s%s", cachedir, ent->d_name);
#if TESTING
                ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                             NULL, "Would remove directory %s", newcachedir);
#else
                rmdir(newcachedir);
#endif
                --nfiles;
            } else {
                /* Directory is not empty. Account for its size: */
                add_long61(&curbytes, ROUNDUP2BLOCKS(buf.st_size));
            }
            continue;
        }
#endif
        /* this is ugly, however I must correspond a header file with a cache file,
           the separate files make writing easier, however GC is nastier */
        if((hdrfile = ((s = strstr(filename, CACHE_HEADER_SUFFIX)) != NULL)) ||
           (s = strstr(filename, CACHE_DATA_SUFFIX)))
        {
            int got_both = 0, baselen = (s - filename);
            apr_cpystrn(basename, filename + basedirlen, (baselen-basedirlen)+1);
        
            for(i = 0; i < files->nelts; i++)
                if(!strcmp(((struct gc_ent *)files->elts + i)->base, basename))
                {
                    got_both = -1; /* this is to flag file I already knew about */
                    break;
                }
            if(got_both == -1) continue;
            
            if(hdrfile)
            {
                 struct stat data_buf;
                char data_file[LARGE_FILE_NAME];
                apr_snprintf(data_file, sizeof(data_file), "%s%s%s", cachebasedir, basename, CACHE_DATA_SUFFIX);
                if(((got_both = stat(data_file, &data_buf) != -1)))
                    buf.st_size += data_buf.st_size;
            }
            else
            {
                struct stat hdr_buf;
                char hdr_file[LARGE_FILE_NAME];
                close(fd);
                apr_snprintf(hdr_file, sizeof(hdr_file), "%s%s%s", cachebasedir, basename, CACHE_HEADER_SUFFIX);
                if((got_both = ((fd = open(hdr_file, O_RDONLY | O_BINARY)) != -1) && fstat(fd, &hdr_buf) != -1))
                    buf.st_size += hdr_buf.st_size;
            }
            if(!got_both)
            {
                ap_log_error(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, r->server,
                             "cache: deleting rogue cache file: %s", filename);
#if TESTING
                ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, NULL,
                             "Would unlink rogue file %s", filename);
#else
                unlink(filename);
#endif
                continue;
            }
        }
        else
        {
            ap_log_error(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, r->server,
                         "cache: unknown cache file: %s", filename);
#if TESTING
            ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                         NULL, "Would unlink unknown file %s", filename);
#else
            unlink(filename);
#endif
            continue;
        }

        /* now read the info */
        i = read(fd, line, 26);
        close(fd);
        if (i == -1)
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
                         "cache gc: read(%s)", filename);
        if (!ap_checkmask(line, "&&&&&&&&&&&&&&&& &&&&&&&&&&&&&&&& &&&&&&&&&&&&&&&&"))
            if((garbage_expire = ap_cache_hex2msec(line + (sizeof(garbage_expire)*2)))== BAD_DATE)
                continue;
            /* bad file */
            if (garbage_now != -1 &&
                (buf.st_atime*APR_USEC_PER_SEC) > garbage_now + MSEC_ONE_DAY &&
                (buf.st_mtime*APR_USEC_PER_SEC) > garbage_now + MSEC_ONE_DAY) {
                ap_log_error(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, r->server,
                             "cache: deleting bad cache file with future date: %s", filename);
#if TESTING
                ap_log_error(APLOG_MARK, APLOG_STARTUP | APLOG_NOERRNO, 0, 
                             NULL, "Would unlink bad file %s", filename);
#else
                unlink(filename);
#endif
                continue;
            }
        /*
         * we need to calculate an 'old' factor, and remove the 'oldest' files
         * so that the space requirement is met; sort by the expires date of the
         * file.
         *
         */
        fent = (struct gc_ent *) apr_push_array(files);
        fent->len = buf.st_size;
        fent->expire = garbage_expire;
        strcpy(fent->base, basename);

    /* accumulate in blocks, to cope with directories > 4Gb */
        add_long61(&curbytes, ROUNDUP2BLOCKS(fent->len));
    }
    closedir(dir);
    return nfiles;
}
